from pathlib import Path

import pytest
import solcx
from ape import Project, reverts
from ape.exceptions import CompilerError
from ape.logging import LogLevel
from ethpm_types import ContractType
from packaging.version import Version

from ape_solidity.exceptions import IndexOutOfBoundsError

EXPECTED_NON_SOLIDITY_ERR_MSG = "Unable to compile 'RandomVyperFile.vy' using Solidity compiler."
raises_because_not_sol = pytest.raises(CompilerError, match=EXPECTED_NON_SOLIDITY_ERR_MSG)


def test_get_config(project, compiler):
    actual = compiler.get_config(project=project)
    assert actual.evm_version == "constantinople"


def test_get_import_remapping(project, compiler):
    actual = compiler.get_import_remapping(project=project)
    expected = {
        "@browniedependency": "contracts/.cache/browniedependency/local",
        "@dependency": "contracts/.cache/dependency/local",
        "@dependencyofdependency": "contracts/.cache/dependencyofdependency/local",
        "@noncompilingdependency": "contracts/.cache/noncompilingdependency/local",
        "@openzeppelin": "contracts/.cache/openzeppelin/4.5.0",
        "@safe": "contracts/.cache/safe/1.3.0",
        "@vault": "contracts/.cache/vault/v0.4.5",
        "@vaultmain": "contracts/.cache/vaultmain/master",
    }
    for key, value in expected.items():
        assert key in actual
        assert actual[key] == value


def test_get_import_remapping_handles_config(project, compiler):
    """
    Show you can override default remappings.
    Normally, these are deduced from dependencies, but you can change them
    and/or add new ones.
    """
    new_value = "NEW_VALUE"
    cfg = {
        "solidity": {
            "import_remapping": [
                "@dependency=dependency",  # Backwards compat!
                "@dependencyofdependency=dependencyofdependency/local",
                f"@vaultmain={new_value}",  # Changing a dependency
                f"@{new_value}={new_value}123",  # Adding something new
            ]
        },
        "dependencies": project.config.dependencies,
    }
    with project.temp_config(**cfg):
        actual = compiler.get_import_remapping(project=project)

    # Show it is backwards compatible (still works w/o changing cfg)
    assert actual["@dependency"] == "contracts/.cache/dependency/local"
    assert actual["@dependencyofdependency"] == "contracts/.cache/dependencyofdependency/local"
    # Show we can change a dependency.
    assert actual["@vaultmain"] == new_value
    # Show we can add a new remapping (quiet dependency).
    assert actual[f"@{new_value}"] == f"{new_value}123"
    # Show other dependency-deduced remappings still work.
    assert actual["@browniedependency"] == "contracts/.cache/browniedependency/local"


def test_get_imports(project, compiler):
    source_id = "contracts/ImportSourceWithEqualSignVersion.sol"
    path = project.sources.lookup(source_id)
    # Source (total) only has these 2 imports.
    expected = (
        "contracts/SpecificVersionWithEqualSign.sol",
        "contracts/CompilesOnce.sol",
    )
    actual = compiler.get_imports((path,), project=project)
    assert source_id in actual
    assert all(e in actual[source_id] for e in expected)


def test_get_imports_indirect(project, compiler):
    """
    Show that twice-removed indirect imports show up. This is required
    for accurate version mapping.
    """

    source_id = "contracts/IndirectlyImportingMoreConstrainedVersion.sol"
    path = project.sources.lookup(source_id)
    expected = (
        # These 2 are directly imported.
        "contracts/ImportSourceWithEqualSignVersion.sol",
        "contracts/IndirectlyImportingMoreConstrainedVersionCompanion.sol",
        # These are 2 are imported by the imported.
        "contracts/SpecificVersionWithEqualSign.sol",
        "contracts/CompilesOnce.sol",
    )
    actual = compiler.get_imports((path,), project=project)
    assert source_id in actual
    actual_str = ", ".join(list(actual[source_id]))
    for ex in expected:
        assert ex in actual[source_id], f"{ex} not in {actual_str}"


def test_get_imports_complex(project, compiler):
    """
    `contracts/Imports.sol` imports sources in every possible
    way. This test shows that we are able to detect all those
    unique ways of importing.
    """
    path = project.sources.lookup("contracts/Imports.sol")
    assert path is not None, "Failed to find Imports test contract."

    actual = compiler.get_imports((path,), project=project)
    expected = {
        "contracts/CompilesOnce.sol": [],
        "contracts/Imports.sol": [
            "contracts/.cache/browniedependency/local/contracts/BrownieContract.sol",
            "contracts/.cache/dependency/local/contracts/Dependency.sol",
            "contracts/.cache/dependencyofdependency/local/contracts/DependencyOfDependency.sol",
            "contracts/.cache/noncompilingdependency/local/contracts/CompilingContract.sol",
            "contracts/.cache/safe/1.3.0/contracts/common/Enum.sol",
            "contracts/CompilesOnce.sol",
            "contracts/MissingPragma.sol",
            "contracts/NumerousDefinitions.sol",
            "contracts/subfolder/Relativecontract.sol",
        ],
        "contracts/MissingPragma.sol": [],
        "contracts/NumerousDefinitions.sol": [],
        "contracts/subfolder/Relativecontract.sol": [],
    }
    for base, imports in expected.items():
        assert base in actual
        assert actual[base] == imports


def test_get_imports_dependencies(project, compiler):
    """
    Show all the affected dependency contracts get included in the imports list.
    """
    source_id = "contracts/UseYearn.sol"
    path = project.sources.lookup(source_id)
    import_ls = compiler.get_imports((path,), project=project)
    actual = import_ls[source_id]
    token_path = "contracts/.cache/openzeppelin/4.5.0/contracts/token"
    expected = [
        f"{token_path}/ERC20/ERC20.sol",
        f"{token_path}/ERC20/IERC20.sol",
        f"{token_path}/ERC20/extensions/IERC20Metadata.sol",
        f"{token_path}/ERC20/utils/SafeERC20.sol",
        "contracts/.cache/openzeppelin/4.5.0/contracts/utils/Address.sol",
        "contracts/.cache/openzeppelin/4.5.0/contracts/utils/Context.sol",
        "contracts/.cache/vault/v0.4.5/contracts/BaseStrategy.sol",
        "contracts/.cache/vaultmain/master/contracts/BaseStrategy.sol",
    ]
    assert actual == expected


def test_get_imports_vyper_file(project, compiler):
    path = Path(__file__).parent / "contracts" / "RandomVyperFile.vy"
    assert path.is_file(), f"Setup failed - file not found {path}"
    with raises_because_not_sol:
        compiler.get_imports((path,))


def test_get_imports_full_project(project, compiler):
    paths = [x for x in project.sources.paths if x.suffix == ".sol"]
    actual = compiler.get_imports(paths, project=project)
    assert len(actual) > 0
    # Prove that every import source also is present in the import map.
    for imported_source_ids in actual.values():
        for source_id in imported_source_ids:
            assert source_id in actual, f"{source_id}'s imports not present."


def test_get_version_map(project, compiler):
    """
    Test that a strict version pragma is recognized in the version map.
    """
    path = project.sources.lookup("contracts/SpecificVersionWithEqualSign.sol")
    actual = compiler.get_version_map((path,), project=project)
    expected_version = Version("0.8.12+commit.f00d7308")
    expected_sources = ("SpecificVersionWithEqualSign",)
    assert expected_version in actual

    actual_ids = [x.stem for x in actual[expected_version]]
    assert all(e in actual_ids for e in expected_sources)


def test_get_version_map_importing_more_constrained_version(project, compiler):
    """
    Test that a strict version pragma in an imported source is recognized
    in the version map.
    """
    # This file's version is not super constrained, but it imports
    # a different source that does have a strict constraint.
    path = project.sources.lookup("contracts/ImportSourceWithEqualSignVersion.sol")

    actual = compiler.get_version_map((path,), project=project)
    expected_version = Version("0.8.12+commit.f00d7308")
    expected_sources = ("ImportSourceWithEqualSignVersion", "SpecificVersionWithEqualSign")
    assert expected_version in actual

    actual_ids = [x.stem for x in actual[expected_version]]
    assert all(e in actual_ids for e in expected_sources)


def test_get_version_map_indirectly_importing_more_constrained_version(project, compiler):
    """
    Test that a strict version pragma in a source imported by an imported
    source (twice removed) is recognized in the version map.
    """
    # This file's version is not super constrained, but it imports
    # a different source that imports another source that does have a constraint.
    path = project.sources.lookup("contracts/IndirectlyImportingMoreConstrainedVersion.sol")

    actual = compiler.get_version_map((path,), project=project)
    expected_version = Version("0.8.12+commit.f00d7308")
    expected_sources = (
        "IndirectlyImportingMoreConstrainedVersion",
        "ImportSourceWithEqualSignVersion",
        "SpecificVersionWithEqualSign",
    )
    assert expected_version in actual

    actual_ids = [x.stem for x in actual[expected_version]]
    assert all(e in actual_ids for e in expected_sources)


def test_get_version_map_dependencies(project, compiler):
    """
    Show all the affected dependency contracts get included in the version map.
    """
    source_id = "contracts/UseYearn.sol"
    older_example = "contracts/ImportOlderDependency.sol"
    paths = [project.sources.lookup(x) for x in (source_id, older_example)]
    actual = compiler.get_version_map(paths, project=project)

    fail_msg = f"versions: {', '.join([str(x) for x in actual])}"
    assert len(actual) == 2, fail_msg

    versions = sorted(list(actual.keys()))
    older = versions[0]  # Via ImportOlderDependency
    latest = versions[1]  # via UseYearn

    oz_token = "contracts/.cache/openzeppelin/4.5.0/contracts/token"
    expected_latest_source_ids = [
        f"{oz_token}/ERC20/ERC20.sol",
        f"{oz_token}/ERC20/IERC20.sol",
        f"{oz_token}/ERC20/extensions/IERC20Metadata.sol",
        f"{oz_token}/ERC20/utils/SafeERC20.sol",
        "contracts/.cache/openzeppelin/4.5.0/contracts/utils/Address.sol",
        "contracts/.cache/openzeppelin/4.5.0/contracts/utils/Context.sol",
        "contracts/.cache/vault/v0.4.5/contracts/BaseStrategy.sol",
        "contracts/.cache/vaultmain/master/contracts/BaseStrategy.sol",
        source_id,
    ]
    expected_older_source_ids = [
        "contracts/.cache/dependency/local/contracts/OlderDependency.sol",
        older_example,
    ]
    expected_latest_source_paths = {project.path / e for e in expected_latest_source_ids}
    expected_oldest_source_paths = {project.path / e for e in expected_older_source_ids}
    assert len(actual[latest]) == len(expected_latest_source_paths)
    assert actual[latest] == expected_latest_source_paths
    assert actual[older] == expected_oldest_source_paths


def test_get_version_map_picks_most_constrained_version(project, compiler):
    """
    Test that if given both a file that can compile at the latest version
    and a file that requires a lesser version but also imports the same file
    that could compile at the latest version, that they are all designated
    to compile using the lesser version.
    """
    source_ids = (
        "contracts/CompilesOnce.sol",
        "contracts/IndirectlyImportingMoreConstrainedVersion.sol",
    )
    paths = [project.sources.lookup(x) for x in source_ids]
    actual = compiler.get_version_map(paths, project=project)
    expected_version = Version("0.8.12+commit.f00d7308")
    assert expected_version in actual
    for path in paths:
        assert path in actual[expected_version], f"{path} is missing!"


def test_get_version_map_version_specified_in_config_file(compiler):
    path = Path(__file__).parent / "VersionSpecifiedInConfig"
    project = Project(path)
    paths = [p for p in project.sources.paths if p.suffix == ".sol"]
    actual = compiler.get_version_map(paths, project=project)
    expected_version = Version("0.8.12+commit.f00d7308")
    assert len(actual) == 1
    assert expected_version in actual
    assert len(actual[expected_version]) > 0


def test_get_version_map_raises_on_non_solidity_sources(project, compiler):
    path = project.contracts_folder / "RandomVyperFile.vy"
    with raises_because_not_sol:
        compiler.get_version_map((path,), project=project)


def test_get_version_map_full_project(project, compiler):
    paths = [x for x in project.sources.paths if x.suffix == ".sol"]
    actual = compiler.get_version_map(paths, project=project)
    latest = sorted(list(actual.keys()), reverse=True)[0]
    v0812 = Version("0.8.12+commit.f00d7308")
    vold = Version("0.4.26+commit.4563c3fc")
    assert v0812 in actual
    assert vold in actual

    v0812_1 = project.path / "contracts/ImportSourceWithEqualSignVersion.sol"
    v0812_2 = project.path / "contracts/IndirectlyImportingMoreConstrainedVersion.sol"

    assert v0812_1 in actual[v0812], "Constrained version files missing"
    assert v0812_2 in actual[v0812], "Constrained version files missing"

    # TDD: This was happening during development of 0.8.0.
    assert v0812_1 not in actual[latest], f"{v0812_1.stem} ended up in latest"
    assert v0812_2 not in actual[latest], f"{v0812_2.stem} ended up in latest"

    # TDD: Old file ending up in multiple spots.
    older_file = project.path / "contracts/.cache/dependency/local/contracts/OlderDependency.sol"
    assert older_file in actual[vold]
    for vers, fileset in actual.items():
        if vers == vold:
            continue

        assert older_file not in fileset, f"Oldest file also appears in version {vers}"


def test_get_compiler_settings(project, compiler):
    path = project.sources.lookup("contracts/Imports.sol")
    actual = compiler.get_compiler_settings((path,), project=project)
    # No reason (when alone) to not use
    assert len(actual) == 1

    # 0.8.12 is hardcoded in some files, but none of those files should be here.
    version = next(iter(actual.keys()))
    assert version > Version("0.8.12+commit.f00d7308")

    settings = actual[version]
    assert settings["optimizer"] == {"enabled": True, "runs": 200}

    # NOTE: These should be sorted!
    assert settings["remappings"] == [
        "@browniedependency=contracts/.cache/browniedependency/local",
        "@dependency=contracts/.cache/dependency/local",
        "@dependencyofdependency=contracts/.cache/dependencyofdependency/local",
        "@noncompilingdependency=contracts/.cache/noncompilingdependency/local",
        "@safe=contracts/.cache/safe/1.3.0",
    ]

    # Set in config.
    assert settings["evmVersion"] == "constantinople"

    # Should be all files (imports of imports etc.)
    actual_files = sorted(list(settings["outputSelection"].keys()))
    expected_files = [
        "contracts/.cache/browniedependency/local/contracts/BrownieContract.sol",
        "contracts/.cache/dependency/local/contracts/Dependency.sol",
        "contracts/.cache/dependencyofdependency/local/contracts/DependencyOfDependency.sol",
        "contracts/.cache/noncompilingdependency/local/contracts/CompilingContract.sol",
        "contracts/.cache/safe/1.3.0/contracts/common/Enum.sol",
        "contracts/CompilesOnce.sol",
        "contracts/Imports.sol",
        "contracts/MissingPragma.sol",
        "contracts/NumerousDefinitions.sol",
        "contracts/subfolder/Relativecontract.sol",
    ]
    assert actual_files == expected_files

    # Output request is the same for all.
    expected_output_request = {
        "*": [
            "abi",
            "bin-runtime",
            "devdoc",
            "userdoc",
            "evm.bytecode.object",
            "evm.bytecode.sourceMap",
            "evm.deployedBytecode.object",
        ],
        "": ["ast"],
    }
    for output in settings["outputSelection"].values():
        assert output == expected_output_request


def test_get_standard_input_json(project, compiler):
    paths = [x for x in project.sources.paths if x.suffix == ".sol"]
    actual = compiler.get_standard_input_json(paths, project=project)
    v0812 = Version("0.8.12+commit.f00d7308")
    v056 = Version("0.5.16+commit.9c3226ce")
    v0426 = Version("0.4.26+commit.4563c3fc")
    latest = sorted(list(actual.keys()), reverse=True)[0]

    fail_msg = f"Versions: {', '.join([str(v) for v in actual])}"
    assert v0812 in actual, fail_msg
    assert v056 in actual, fail_msg
    assert v0426 in actual, fail_msg
    assert latest in actual, fail_msg

    v0812_sources = list(actual[v0812]["sources"].keys())
    v056_sources = list(actual[v056]["sources"].keys())
    v0426_sources = list(actual[v0426]["sources"].keys())
    latest_sources = list(actual[latest]["sources"].keys())

    assert "contracts/ImportSourceWithEqualSignVersion.sol" not in latest_sources
    assert "contracts/IndirectlyImportingMoreConstrainedVersion.sol" not in latest_sources

    # Some source expectations.
    assert "contracts/CompilesOnce.sol" in v0812_sources
    assert "contracts/SpecificVersionRange.sol" in v0812_sources
    assert "contracts/ImportSourceWithNoPrefixVersion.sol" in v0812_sources

    assert "contracts/OlderVersion.sol" in v056_sources
    assert "contracts/ImportOlderDependency.sol" in v0426_sources
    assert "contracts/.cache/dependency/local/contracts/OlderDependency.sol" in v0426_sources


def test_compile(project, compiler):
    path = project.sources.lookup("contracts/Imports.sol")
    actual = [c for c in compiler.compile((path,), project=project)]
    # We only get back the contracts we requested, even if it had to compile
    # others (like imports) to get it to work.
    assert len(actual) == 1
    assert isinstance(actual[0], ContractType)
    assert actual[0].name == "Imports"
    assert actual[0].source_id == "contracts/Imports.sol"
    assert actual[0].deployment_bytecode is not None
    assert actual[0].runtime_bytecode is not None
    assert len(actual[0].abi) > 0


def test_compile_performance(benchmark, compiler, project):
    """
    See https://pytest-benchmark.readthedocs.io/en/latest/
    """
    path = project.sources.lookup("contracts/MultipleDefinitions.sol")
    result = benchmark.pedantic(
        lambda *args, **kwargs: [x for x in compiler.compile(*args, **kwargs)],
        args=((path,),),
        kwargs={"project": project},
        rounds=1,
    )
    assert len(result) > 0


def test_compile_multiple_definitions_in_source(project, compiler):
    """
    Show that if multiple contracts / interfaces are defined in a single
    source, that we get all of them when compiling.
    """
    source_id = "contracts/MultipleDefinitions.sol"
    path = project.sources.lookup(source_id)
    result = [c for c in compiler.compile((path,), project=project)]
    assert len(result) == 2
    assert [r.name for r in result] == ["IMultipleDefinitions", "MultipleDefinitions"]
    assert all(r.source_id == source_id for r in result)

    assert project.MultipleDefinitions
    assert project.IMultipleDefinitions


def test_compile_contract_with_different_name_than_file(project, compiler):
    source_id = "contracts/DifferentNameThanFile.sol"
    path = project.sources.lookup(source_id)
    actual = [c for c in compiler.compile((path,), project=project)]
    assert len(actual) == 1
    assert actual[0].source_id == source_id


def test_compile_only_returns_contract_types_for_inputs(project, compiler):
    """
    Test showing only the requested contract types get returned.
    """
    path = project.sources.lookup("contracts/Imports.sol")
    contract_types = [c for c in compiler.compile((path,), project=project)]
    assert len(contract_types) == 1
    assert contract_types[0].name == "Imports"


def test_compile_vyper_contract(project, compiler):
    path = project.contracts_folder / "RandomVyperFile.vy"
    with raises_because_not_sol:
        _ = [c for c in compiler.compile((path,), project=project)]


def test_compile_just_a_struct(compiler, project):
    """
    Before, you would get a nasty index error, even though this is valid Solidity.
    The fix involved using nicer access to "contracts" in the standard output JSON.
    """
    path = project.sources.lookup("contracts/JustAStruct.sol")
    contract_types = [c for c in compiler.compile((path,), project=project)]
    assert len(contract_types) == 0


def test_compile_produces_source_map(project, compiler):
    path = project.sources.lookup("contracts/MultipleDefinitions.sol")
    result = [c for c in compiler.compile((path,), project=project)][-1]
    assert result.sourcemap.root == "124:87:0:-:0;;;;;;;;;;;;;;;;;;;"


def test_compile_via_ir(project, compiler):
    path = project.contracts_folder / "StackTooDep.sol"
    source_code = """
// SPDX-License-Identifier: MIT

pragma solidity >=0.8.0;

contract StackTooDeep {
    // This contract tests the scenario when we have a contract with
    // too many local variables and the stack is too deep.
    // The compiler will throw an error when trying to compile this contract.
    // To get around the error, we can compile the contract with the
    // --via-ir flag

    function foo(
        uint256 a,
        uint256 b,
        uint256 c,
        uint256 d,
        uint256 e,
        uint256 f,
        uint256 g,
        uint256 h,
        uint256 i,
        uint256 j,
        uint256 k,
        uint256 l,
        uint256 m,
        uint256 n,
        uint256 o,
        uint256 p
    ) public pure returns (uint256) {

        uint256 sum = 0;

        for (uint256 index = 0; index < 16; index++) {
            uint256 innerSum = a + b + c + d + e + f + g + h + i + j + k + l + m + n + o + p;
            sum += innerSum;
        }

        return (sum);
    }

}
    """

    # write source code to file
    path.write_text(source_code)

    try:
        [c for c in compiler.compile((path,), project=project)]
    except Exception as e:
        assert "Stack too deep" in str(e)

    with project.temp_config(solidity={"via_ir": True}):
        _ = [c for c in compiler.compile((path,), project=project)]

        # delete source code file
        path.unlink()


@pytest.mark.install
def test_installs_from_compile(project, compiler, temp_solcx_path):
    """
    Test the compilation of a contract with no defined pragma spec.

    The plugin should implicitly download the latest version to compile the
    contract with. `temp_solcx_path` is used to simulate an environment without
    compilers installed.
    """
    assert not solcx.get_installed_solc_versions()
    path = project.sources.lookup("contracts/MissingPragma.sol")
    contract_types = [c for c in compiler.compile((path,), project=project)]
    assert len(contract_types) == 1
    installed_versions = solcx.get_installed_solc_versions()
    assert len(installed_versions) == 1
    assert installed_versions[0] == max(solcx.get_installable_solc_versions())


def test_compile_project(project, compiler):
    """
    Simple test showing the full project indeed compiles.
    """
    paths = [x for x in project.sources.paths if x.suffix == ".sol"]
    actual = [c for c in compiler.compile(paths, project=project)]
    assert len(actual) > 0


def test_compile_outputs_compiler_data_to_manifest(project, compiler):
    project.update_manifest(compilers=[])
    path = project.sources.lookup("contracts/CompilesOnce.sol")
    _ = [c for c in compiler.compile((path,), project=project)]
    assert len(project.manifest.compilers or []) == 1
    actual = project.manifest.compilers[0]
    assert actual.name == "solidity"
    assert "CompilesOnce" in actual.contractTypes
    assert actual.version == "0.8.26+commit.8a97fa7a"
    # Compiling again should not add the same compiler again.
    _ = [c for c in compiler.compile((path,), project=project)]
    length_again = len(project.manifest.compilers or [])
    assert length_again == 1


def test_add_library(project, account, compiler, connection):
    # Does not exist yet because library is not deployed or known.
    with pytest.raises(AttributeError):
        _ = project.ContractUsingLibraryInSameSource
    with pytest.raises(AttributeError):
        _ = project.ContractUsingLibraryNotInSameSource

    library = project.ExampleLibrary.deploy(sender=account)
    compiler.add_library(library, project=project)

    # After deploying and adding the library, we can use contracts that need it.
    assert project.ContractUsingLibraryNotInSameSource
    assert project.ContractUsingLibraryInSameSource


def test_enrich_error_when_custom(compiler, project, owner, not_owner, connection):
    path = project.sources.lookup("contracts/HasError.sol")
    _ = [c for c in compiler.compile((path,), project=project)]

    # Deploy so Ape know about contract type.
    contract = owner.deploy(project.HasError, 1)
    with pytest.raises(contract.Unauthorized) as err:
        contract.withdraw(sender=not_owner)

    assert err.value.inputs == {"addr": not_owner.address, "counter": 123}


def test_enrich_error_when_custom_in_constructor(compiler, project, owner, not_owner, connection):
    # Deploy so Ape know about contract type.
    with reverts(project.HasError.Unauthorized) as err:
        not_owner.deploy(project.HasError, 0)

    assert err.value.inputs == {"addr": not_owner.address, "counter": 123}


def test_enrich_error_when_builtin(project, owner, connection):
    contract = project.BuiltinErrorChecker.deploy(sender=owner)
    with pytest.raises(IndexOutOfBoundsError):
        contract.checkIndexOutOfBounds(sender=owner)


def test_flatten(project, compiler, caplog):
    path = project.sources.lookup("contracts/Imports.sol")
    with caplog.at_level(LogLevel.WARNING):
        compiler.flatten_contract(path, project=project)
        actual = caplog.messages[-1]
        expected = (
            "Conflicting licenses found: 'LGPL-3.0-only, MIT'. "
            "Using the root file's license 'MIT'."
        )
        assert actual == expected

    path = project.sources.lookup("contracts/ImportingLessConstrainedVersion.sol")
    flattened_source = compiler.flatten_contract(path, project=project)
    flattened_source_path = (
        Path(__file__).parent / "data" / "ImportingLessConstrainedVersionFlat.sol"
    )
    assert str(flattened_source) == str(flattened_source_path.read_text())


def test_compile_code(project, compiler):
    code = """
contract Contract {
    function snakes() pure public returns(bool) {
        return true;
    }
}
"""
    actual = compiler.compile_code(code, project=project, contractName="TestContractName")
    assert actual.name == "TestContractName"
    assert len(actual.abi) > 0
    assert actual.ast is not None
    assert len(actual.runtime_bytecode.bytecode) > 0
    assert len(actual.deployment_bytecode.bytecode) > 0



import pytest
from ape.logging import logger
from solcx.exceptions import SolcError

from ape_solidity.exceptions import SolcCompileError

MESSAGE = "__message__"
COMMAND = ["solc", "command"]
RETURN_CODE = 123
STDOUT_DATA = "<stdout data>"
STDERR_DATA = "<stderr data>"


@pytest.fixture(scope="module")
def solc_error():
    return SolcError(
        message=MESSAGE,
        command=COMMAND,
        return_code=RETURN_CODE,
        stdout_data=STDOUT_DATA,
        stderr_data=STDERR_DATA,
    )


def test_solc_compile_error(solc_error):
    error = SolcCompileError(solc_error)
    actual = str(error)
    assert MESSAGE in actual
    assert f"{RETURN_CODE}" not in actual
    assert " ".join(COMMAND) not in actual
    assert STDOUT_DATA not in actual
    assert STDERR_DATA not in actual


def test_solc_compile_error_verbose(solc_error):
    logger.set_level("DEBUG")
    error = SolcCompileError(solc_error)
    actual = str(error)
    assert MESSAGE in actual
    assert f"{RETURN_CODE}" in actual
    assert " ".join(COMMAND) in actual
    assert STDOUT_DATA in actual
    assert STDERR_DATA in actual
    logger.set_level("INFO")



import shutil
from contextlib import contextmanager
from pathlib import Path
from tempfile import mkdtemp
from unittest import mock

import ape
import pytest
import solcx
from ape.utils.os import create_tempdir

from ape_solidity._utils import Extension


@contextmanager
def _tmp_solcx_path(monkeypatch):
    solcx_install_path = mkdtemp()

    monkeypatch.setenv(
        solcx.install.SOLCX_BINARY_PATH_VARIABLE,
        solcx_install_path,
    )

    yield solcx_install_path

    if Path(solcx_install_path).is_dir():
        shutil.rmtree(solcx_install_path, ignore_errors=True)


@pytest.fixture
def fake_no_installs(mocker):
    """
    Tricks the tests into thinking there are no installed versions.
    This saves time because it won't actually need to install solc,
    and it should still work.
    """
    patch = mocker.patch("ape_solidity.compiler.get_installed_solc_versions")
    patch.return_value = []
    return patch


@pytest.fixture
def temp_solcx_path(monkeypatch):
    """
    Creates a new, temporary installation path for solcx for a given test.
    """
    with _tmp_solcx_path(monkeypatch) as path:
        yield path


@pytest.fixture(scope="session")
def project(config):
    _ = config  # Ensure temp data folder gets set first.
    root = Path(__file__).parent

    # Delete .build / .cache that may exist pre-copy
    for path in (
        root,
        root / "BrownieProject",
        root / "BrownieStyleDependency",
        root / "Dependency",
        root / "DependencyOfDependency",
        root / "ProjectWithinProject",
        root / "VersionSpecifiedInConfig",
    ):
        for cache in (path / ".build", path / "contracts" / ".cache"):
            if cache.is_dir():
                shutil.rmtree(cache)

    root_project = ape.Project(root)
    with root_project.isolate_in_tempdir() as tmp_project:
        yield tmp_project


@pytest.fixture(scope="session", autouse=True)
def config():
    cfg = ape.config

    # Uncomment to install dependencies in actual data folder.
    # This will save time running tests.
    # project = ape.Project(Path(__file__).parent)
    # project.dependencies.install()

    # Ensure we don't persist any .ape data.
    real_data_folder = cfg.DATA_FOLDER
    with create_tempdir() as path:
        cfg.DATA_FOLDER = path

        # Copy in existing packages to save test time
        # when running locally.
        packages = real_data_folder / "packages"
        packages.mkdir(parents=True, exist_ok=True)
        shutil.copytree(packages, path / "packages", dirs_exist_ok=True)

        yield cfg


@pytest.fixture
def compiler_manager():
    return ape.compilers


@pytest.fixture
def compiler(compiler_manager):
    return compiler_manager.solidity


@pytest.fixture(autouse=True)
def ignore_other_compilers(mocker, compiler_manager, compiler):
    """
    Having ape-vyper installed causes the random Vyper file
    (that exists for testing purposes) to get compiled and
    vyper to get repeatedly installed in a temporary directory.
    Avoid that by tricking Ape into thinking ape-vyper is not
    installed (if it is).
    """
    existing_compilers = compiler_manager.registered_compilers
    ape_pm = compiler_manager.ethpm
    valid_compilers = {
        ext: c for ext, c in existing_compilers.items() if ext in [x.value for x in Extension]
    }
    path = "ape.managers.compilers.CompilerManager.registered_compilers"
    mock_registered_compilers = mocker.patch(path, new_callable=mock.PropertyMock)

    # Only ethpm (.json) and Solidity extensions allowed.
    mock_registered_compilers.return_value = {".json": ape_pm, **valid_compilers}


@pytest.fixture
def account():
    return ape.accounts.test_accounts[0]


@pytest.fixture
def owner():
    return ape.accounts.test_accounts[1]


@pytest.fixture
def not_owner():
    return ape.accounts.test_accounts[2]


@pytest.fixture
def connection():
    with ape.networks.ethereum.local.use_provider("test") as provider:
        yield provider



from enum import IntEnum
from typing import Union

from ape.exceptions import CompilerError, ConfigError, ContractLogicError
from ape.logging import LogLevel, logger
from solcx.exceptions import SolcError


class SolcInstallError(CompilerError):
    """
    Raised when `solc` is not installed
    and unable to be installed.
    """

    def __init__(self):
        super().__init__(
            "No versions of `solc` installed and unable to install latest `solc` version."
        )


class SolcCompileError(CompilerError):
    """
    Specifically, only errors arising from ``solc`` compile methods.
    This error is a modified version of ``SolcError`` to take into
    account Ape's logging verbosity.
    """

    def __init__(self, solc_error: SolcError):
        self.solc_error = solc_error

    def __str__(self) -> str:
        if logger.level <= LogLevel.DEBUG:
            # Show everything when in DEBUG mode.
            return str(self.solc_error)

        else:
            # Only show the error and line-number(s) where it occurred.
            return self.solc_error.message


class IncorrectMappingFormatError(ConfigError, ValueError):
    def __init__(self):
        super().__init__(
            "Incorrectly formatted 'solidity.remapping' config property. "
            "Expected '@value_1=value2'."
        )


class RuntimeErrorType(IntEnum):
    ASSERTION_ERROR = 0x1
    ARITHMETIC_UNDER_OR_OVERFLOW = 0x11
    DIVISION_BY_ZERO_ERROR = 0x12
    ENUM_CONVERSION_OUT_OF_BOUNDS = 0x21
    INCORRECTLY_ENCODED_STORAGE_BYTE_ARRAY = 0x22
    POP_ON_EMPTY_ARRAY = 0x31
    INDEX_OUT_OF_BOUNDS_ERROR = 0x32
    MEMORY_OVERFLOW_ERROR = 0x41
    ZERO_INITIALIZED_VARIABLE = 0x51


class SolidityRuntimeError(ContractLogicError):
    def __init__(self, error_type: RuntimeErrorType, message: str, **kwargs):
        self.error_type = error_type
        super().__init__(message, **kwargs)


class SolidityArithmeticError(SolidityRuntimeError, ArithmeticError):
    """
    Raised from math operations going wrong.
    """

    def __init__(self, **kwargs):
        message = "Arithmetic operation underflowed or overflowed outside of an unchecked block."
        super().__init__(RuntimeErrorType.ARITHMETIC_UNDER_OR_OVERFLOW, message, **kwargs)


class SolidityAssertionError(SolidityRuntimeError, AssertionError):
    """
    Raised from Solidity ``assert`` statements.
    You typically should never see this error, as higher-level Contract Logic error
    handled in the framework should appear first (with the correct revert message).
    """

    def __init__(self, **kwargs):
        message = "Assertion error."
        super().__init__(RuntimeErrorType.ASSERTION_ERROR, message, **kwargs)


class DivisionByZeroError(SolidityRuntimeError, ZeroDivisionError):
    """
    Raised when dividing goes wrong (such as using a 0 denominator).
    """

    def __init__(self, **kwargs):
        message = "Division or modulo division by zero"
        super().__init__(RuntimeErrorType.DIVISION_BY_ZERO_ERROR, message, **kwargs)


class EnumConversionError(SolidityRuntimeError):
    """
    Raised when Solidity fails to convert an enum value to its primitive type.
    """

    def __init__(self, **kwargs):
        message = "Tried to convert a value into an enum, but the value was too big or negative."
        super().__init__(RuntimeErrorType.ENUM_CONVERSION_OUT_OF_BOUNDS, message, **kwargs)


class EncodeStorageError(SolidityRuntimeError):
    """
    Raised when Solidity fails to encode a storage value.
    """

    def __init__(self, **kwargs):
        message = "Incorrectly encoded storage byte array."
        super().__init__(RuntimeErrorType.INCORRECTLY_ENCODED_STORAGE_BYTE_ARRAY, message, **kwargs)


class IndexOutOfBoundsError(SolidityRuntimeError, IndexError):
    """
    Raised when accessing an index that is out of bounds in your contract.
    """

    def __init__(self, **kwargs):
        message = "Array accessed at an out-of-bounds or negative index."
        super().__init__(RuntimeErrorType.INDEX_OUT_OF_BOUNDS_ERROR, message, **kwargs)


class MemoryOverflowError(SolidityRuntimeError, OverflowError):
    """
    Raised when exceeding the allocating memory for a data type
    in Solidity.
    """

    def __init__(self, **kwargs):
        message = "Too much memory was allocated, or an array was created that is too large."
        super().__init__(RuntimeErrorType.MEMORY_OVERFLOW_ERROR, message, **kwargs)


class PopOnEmptyArrayError(SolidityRuntimeError):
    """
    Raised when popping from a data-structure fails in your contract.
    """

    def __init__(self, **kwargs):
        message = ".pop() was called on an empty array."
        super().__init__(RuntimeErrorType.POP_ON_EMPTY_ARRAY, message, **kwargs)


class ZeroInitializedVariableError(SolidityRuntimeError):
    """
    Raised when calling a zero-initialized variable of internal function type.
    """

    def __init__(self, **kwargs):
        message = "Called a zero-initialized variable of internal function type."
        super().__init__(RuntimeErrorType.ZERO_INITIALIZED_VARIABLE, message, **kwargs)


RUNTIME_ERROR_CODE_PREFIX = "0x4e487b71"
RuntimeErrorUnion = Union[
    SolidityArithmeticError,
    SolidityAssertionError,
    DivisionByZeroError,
    EnumConversionError,
    EncodeStorageError,
    IndexOutOfBoundsError,
    MemoryOverflowError,
    PopOnEmptyArrayError,
    ZeroInitializedVariableError,
]
RUNTIME_ERROR_MAP: dict[RuntimeErrorType, type[RuntimeErrorUnion]] = {
    RuntimeErrorType.ASSERTION_ERROR: SolidityAssertionError,
    RuntimeErrorType.ARITHMETIC_UNDER_OR_OVERFLOW: SolidityArithmeticError,
    RuntimeErrorType.DIVISION_BY_ZERO_ERROR: DivisionByZeroError,
    RuntimeErrorType.ENUM_CONVERSION_OUT_OF_BOUNDS: EnumConversionError,
    RuntimeErrorType.INCORRECTLY_ENCODED_STORAGE_BYTE_ARRAY: EncodeStorageError,
    RuntimeErrorType.INDEX_OUT_OF_BOUNDS_ERROR: IndexOutOfBoundsError,
    RuntimeErrorType.MEMORY_OVERFLOW_ERROR: MemoryOverflowError,
    RuntimeErrorType.POP_ON_EMPTY_ARRAY: PopOnEmptyArrayError,
    RuntimeErrorType.ZERO_INITIALIZED_VARIABLE: ZeroInitializedVariableError,
}



import pytest
from ape._cli import cli
from click.testing import CliRunner


@pytest.fixture
def ape_cli():
    return cli


@pytest.fixture
def runner():
    return CliRunner()


def test_compile_using_cli(ape_cli, runner, project):
    arguments = ["compile", "--project", f"{project.path}"]
    result = runner.invoke(ape_cli, [*arguments, "--force"], catch_exceptions=False)
    assert result.exit_code == 0
    assert "CompilesOnce" in result.output
    result = runner.invoke(ape_cli, arguments, catch_exceptions=False)

    # Already compiled so does not compile again.
    assert "CompilesOnce" not in result.output


@pytest.mark.parametrize(
    "contract_path",
    (
        "CompilesOnce",
        "CompilesOnce.sol",
        "contracts/CompilesOnce",
        "contracts/CompilesOnce.sol",
    ),
)
def test_compile_specified_contracts(ape_cli, runner, contract_path, project):
    arguments = ("compile", "--project", f"{project.path}", contract_path, "--force")
    result = runner.invoke(ape_cli, arguments, catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert "contracts/CompilesOnce.sol" in result.output, f"Failed to compile {contract_path}."



import json
import re
from collections.abc import Iterable
from enum import Enum
from pathlib import Path
from typing import Optional, Union

from ape.exceptions import CompilerError
from ape.utils import pragma_str_to_specifier_set
from packaging.specifiers import SpecifierSet
from packaging.version import Version
from solcx.install import get_executable
from solcx.wrapper import get_solc_version as get_solc_version_from_binary

OUTPUT_SELECTION = [
    "abi",
    "bin-runtime",
    "devdoc",
    "userdoc",
    "evm.bytecode.object",
    "evm.bytecode.sourceMap",
    "evm.deployedBytecode.object",
]


class Extension(Enum):
    SOL = ".sol"


def get_import_lines(source_paths: Iterable[Path]) -> dict[Path, list[str]]:
    imports_dict: dict[Path, list[str]] = {}
    for filepath in source_paths:
        import_set = set()
        if not filepath or not filepath.is_file():
            continue

        source_lines = filepath.read_text().splitlines()
        num_lines = len(source_lines)
        for line_number, ln in enumerate(source_lines):
            if not ln.startswith("import"):
                continue

            import_str = ln
            second_line_number = line_number
            while ";" not in import_str:
                second_line_number += 1
                if second_line_number >= num_lines:
                    raise CompilerError("Import statement missing semicolon.")

                next_line = source_lines[second_line_number]
                import_str += f" {next_line.strip()}"

            import_set.add(import_str)
            line_number += 1

        imports_dict[filepath] = list(import_set)

    return imports_dict


def get_pragma_spec_from_path(source_file_path: Union[Path, str]) -> Optional[SpecifierSet]:
    """
    Extracts pragma information from Solidity source code.

    Args:
        source_file_path (Union[Path, str]): Solidity source file path.

    Returns:
        ``Optional[packaging.specifiers.SpecifierSet]``
    """
    path = Path(source_file_path)
    if not path.is_file():
        return None

    source_str = path.read_text()
    return get_pragma_spec_from_str(source_str)


def get_pragma_spec_from_str(source_str: str) -> Optional[SpecifierSet]:
    if not (
        pragma_match := next(
            re.finditer(r"(?:\n|^)\s*pragma\s*solidity\s*([^;\n]*)", source_str), None
        )
    ):
        return None  # Try compiling with latest

    return pragma_str_to_specifier_set(pragma_match.groups()[0])


def load_dict(data: Union[str, dict]) -> dict:
    return data if isinstance(data, dict) else json.loads(data)


def add_commit_hash(version: Union[str, Version]) -> Version:
    vers = Version(f"{version}") if isinstance(version, str) else version
    has_commit = len(f"{vers}") > len(vers.base_version)
    if has_commit:
        # Already added.
        return vers

    solc = get_executable(version=vers)
    return get_solc_version_from_binary(solc, with_commit_hash=True)


def get_versions_can_use(pragma_spec: SpecifierSet, options: Iterable[Version]) -> list[Version]:
    return sorted(list(pragma_spec.filter(options)), reverse=True)


def select_version(pragma_spec: SpecifierSet, options: Iterable[Version]) -> Optional[Version]:
    choices = get_versions_can_use(pragma_spec, options)
    return choices[0] if choices else None


def strip_commit_hash(version: Union[str, Version]) -> Version:
    """
    Version('0.8.21+commit.d9974bed') => Version('0.8.21')> the simple way.
    """
    return Version(f"{str(version).split('+')[0].strip()}")



import os
import re
from collections.abc import Iterable, Iterator
from pathlib import Path
from typing import Any, Optional, Union

from ape.api import CompilerAPI, PluginConfig
from ape.contracts import ContractInstance
from ape.exceptions import CompilerError, ConfigError, ContractLogicError, ProjectError
from ape.logging import logger
from ape.managers.project import ProjectManager
from ape.types import AddressType, ContractType
from ape.utils import cached_property, get_full_extension, get_relative_path
from ape.version import version
from eth_pydantic_types import HexBytes
from eth_utils import add_0x_prefix, is_0x_prefixed
from ethpm_types.source import Compiler, Content
from packaging.specifiers import SpecifierSet
from packaging.version import Version
from pydantic import model_validator
from requests.exceptions import ConnectionError
from solcx import (
    compile_source,
    compile_standard,
    get_installable_solc_versions,
    get_installed_solc_versions,
    install_solc,
)
from solcx.exceptions import SolcError
from solcx.install import get_executable

from ape_solidity._utils import (
    OUTPUT_SELECTION,
    Extension,
    add_commit_hash,
    get_import_lines,
    get_pragma_spec_from_path,
    get_pragma_spec_from_str,
    get_versions_can_use,
    load_dict,
    select_version,
    strip_commit_hash,
)
from ape_solidity.exceptions import (
    RUNTIME_ERROR_CODE_PREFIX,
    RUNTIME_ERROR_MAP,
    RuntimeErrorType,
    RuntimeErrorUnion,
    SolcCompileError,
    SolcInstallError,
)

# Define a regex pattern that matches import statements
# Both single and multi-line imports will be matched
IMPORTS_PATTERN = re.compile(
    r"import\s+((.*?)(?=;)|[\s\S]*?from\s+(.*?)(?=;));\s", flags=re.MULTILINE
)
LICENSES_PATTERN = re.compile(r"(// SPDX-License-Identifier:\s*([^\n]*)\s)")
VERSION_PRAGMA_PATTERN = re.compile(r"pragma solidity[^;]*;")
DEFAULT_OPTIMIZATION_RUNS = 200


class ImportRemapping(PluginConfig):
    """
    A remapped import set in the config.
    """

    @model_validator(mode="before")
    def validate_str(cls, value):
        if isinstance(value, str):
            parts = value.split("=")
            return {"key": parts[0], "value": parts[1]}

        return value

    """
    The key of the remapping, such as ``@openzeppelin``.
    """
    key: str

    """
    The value to use in place of the key,
    such as ``path/somewhere/else``.
    """
    value: str

    def __str__(self) -> str:
        return f"{self.key}={self.value}"

    def __eq__(self, other):
        if isinstance(other, str):
            return str(self) == other

        return super().__eq__(other)


class SolidityConfig(PluginConfig):
    """
    Configure the ape-solidity plugin.
    """

    import_remapping: list[ImportRemapping] = []
    """
    Custom remappings as a key value map..
    Note: You do not need to specify dependencies here.
    """

    optimize: bool = True
    """
    Compile with optimization. Defaults to ``True``.
    """

    version: Optional[str] = None
    """
    Hardcode a Solidity version to use. When not set,
    ape-solidity attempts to use the best version(s)
    available.
    """

    evm_version: Optional[str] = None
    """
    Compile targeting this EVM version.
    """

    via_ir: Optional[bool] = None
    """
    Set to ``True`` to turn on compilation mode via the IR.
    Defaults to ``None`` which does not pass the flag to
    the compiler (same as ``False``).
    """


def _get_flattened_source(path: Path, name: Optional[str] = None) -> str:
    name = name or path.name
    result = f"// File: {name}\n"
    result += path.read_text() + "\n"
    return result


class SolidityCompiler(CompilerAPI):
    """
    The implementation of the ape-solidity Compiler class.
    Implements all methods in :class:`~ape.api.compilers.CompilerAPI`.
    Compiles ``.sol`` files into ``ContractTypes`` for usage in the
    Ape framework.
    """

    # Libraries adding for linking. See `add_library` method.
    _libraries: dict[str, dict[str, AddressType]] = {}

    @property
    def name(self) -> str:
        return "solidity"

    @property
    def libraries(self) -> dict[str, dict[str, AddressType]]:
        return self._libraries

    @cached_property
    def available_versions(self) -> list[Version]:
        try:
            return get_installable_solc_versions()
        except ConnectionError:
            # Compiling offline
            logger.warning("Internet connection required to fetch installable Solidity versions.")
            return []

    @property
    def installed_versions(self) -> list[Version]:
        """
        Returns a lis of installed version WITHOUT their
        commit hashes.
        """
        return get_installed_solc_versions()

    @property
    def latest_version(self) -> Optional[Version]:
        """
        Returns the latest version available of ``solc``.
        When unable to retrieve available ``solc`` versions, such as
        times disconnected from the Internet, returns ``None``.
        """
        return _try_max(self.available_versions)

    @property
    def latest_installed_version(self) -> Optional[Version]:
        """
        Returns the highest version of all the installed versions.
        If ``solc`` is not installed at all, returns ``None``.
        """
        return _try_max(self.installed_versions)

    def _get_configured_version(
        self, project: Optional[ProjectManager] = None
    ) -> Optional[Version]:
        """
        A helper property that gets, verifies, and installs (if needed)
        the version specified in the config.
        """
        pm = project or self.local_project
        config = self.get_config(project=pm)
        if not (version := config.version):
            return None

        installed_versions = self.installed_versions
        specified_commit_hash = "+" in version
        base_version = strip_commit_hash(version)
        if base_version not in installed_versions:
            install_solc(base_version, show_progress=True)

        settings_version = add_commit_hash(base_version)
        if specified_commit_hash and settings_version != version:
            raise ConfigError(
                f"Commit hash from settings version {version} "
                f"differs from installed: {settings_version}"
            )

        return settings_version

    @cached_property
    def _ape_version(self) -> Version:
        return Version(version.split(".dev")[0].strip())

    def add_library(self, *contracts: ContractInstance, project: Optional[ProjectManager] = None):
        """
        Set a library contract type address. This is useful when deploying a library
        in a local network and then adding the address afterward. Now, when
        compiling again, it will use the new address.

        Args:
            *contracts (``ContractInstance``): The deployed library contract(s).
            project (Optional[ProjectManager]): The project using the library.
        """
        pm = project or self.local_project
        for contract in contracts:
            if not (source_id := contract.contract_type.source_id):
                raise CompilerError("Missing source ID.")
            elif not (name := contract.contract_type.name):
                raise CompilerError("Missing contract type name.")

            self._libraries[source_id] = {name: contract.address}
            path = pm.path / source_id
            if not path.is_file():
                return

            # Recompile the same source, in case contracts were in there
            # that required the libraries.
            contract_types = {
                ct.name: ct for ct in self.compile((path,), project=project) if ct.name
            }
            if contract_types:
                all_types = {**pm.manifest.contract_types, **contract_types}
                pm.update_manifest(contract_types=all_types)

    def get_versions(self, all_paths: Iterable[Path]) -> set[str]:
        _validate_can_compile(all_paths)
        versions = set()
        for path in all_paths:
            # Make sure we have the compiler available to compile this
            if version_spec := get_pragma_spec_from_path(path):
                if selected_version := select_version(version_spec, self.available_versions):
                    versions.add(selected_version.base_version)

        return versions

    def get_import_remapping(self, project: Optional[ProjectManager] = None) -> dict[str, str]:
        """
        Config remappings like ``'@import_name=path/to/dependency'`` parsed here
        as ``{'@import_name': 'path/to/dependency'}``.

        Returns:
            Dict[str, str]: Where the key is the import name, e.g. ``"@openzeppelin"`
            and the value is a stringified relative path (source ID) of the cached contract,
            e.g. `".cache/openzeppelin/4.4.2".
        """
        pm = project or self.local_project
        prefix = f"{get_relative_path(pm.contracts_folder, pm.path)}"

        specified = pm.dependencies.install()

        # Ensure .cache folder is ready-to-go.
        cache_folder = pm.contracts_folder / ".cache"
        cache_folder.mkdir(exist_ok=True, parents=True)

        # Start with explicitly configured remappings.
        cfg_remappings: dict[str, str] = {
            m.key: m.value for m in pm.config.solidity.import_remapping
        }
        key_map: dict[str, str] = {}

        def get_cache_id(dep) -> str:
            return os.path.sep.join((prefix, ".cache", dep.name, dep.version))

        def unpack(dep):
            # Ensure the dependency is installed.
            try:
                dep.project
            except ProjectError:
                # Try to compile anyway.
                # Let the compiler fail on its own.
                return

            for unpacked_dep in dep.unpack(pm.contracts_folder / ".cache"):
                _key = key_map.get(unpacked_dep.name, f"@{unpacked_dep.name}")
                if _key not in remapping:
                    remapping[_key] = get_cache_id(unpacked_dep)
                # else, was specified or configured more appropriately.

        remapping: dict[str, str] = {}
        for key, value in cfg_remappings.items():
            # Check if legacy-style and still accept it.
            parts = value.split(os.path.sep)
            name = parts[0]
            _version = None
            if len(parts) > 2:
                # Clearly, not pointing at a dependency.
                remapping[key] = value
                continue

            elif len(parts) == 2:
                _version = parts[1]

            if _version is None:
                matching_deps = [d for d in pm.dependencies.installed if d.name == name]
                if len(matching_deps) == 1:
                    _version = matching_deps[0].version
                else:
                    # Not obvious if it is pointing at one of these dependencies.
                    remapping[key] = value
                    continue

            # Dependency found. Map to it using the provider key.
            dependency = pm.dependencies.get_dependency(name, _version)
            key_map[dependency.name] = key
            unpack(dependency)

        # Add auto-remapped dependencies.
        # (Meaning, the dependencies are specified but their remappings
        # are not, so we auto-generate default ones).
        for dependency in specified:
            unpack(dependency)

        return remapping

    def get_compiler_settings(
        self, contract_filepaths: Iterable[Path], project: Optional[ProjectManager] = None, **kwargs
    ) -> dict[Version, dict]:
        pm = project or self.local_project
        _validate_can_compile(contract_filepaths)
        remapping = self.get_import_remapping(project=pm)
        imports = self.get_imports_from_remapping(contract_filepaths, remapping, project=pm)
        return self._get_settings_from_imports(contract_filepaths, imports, remapping, project=pm)

    def _get_settings_from_imports(
        self,
        contract_filepaths: Iterable[Path],
        import_map: dict[str, list[str]],
        remappings: dict[str, str],
        project: Optional[ProjectManager] = None,
    ):
        pm = project or self.local_project
        files_by_solc_version = self.get_version_map_from_imports(
            contract_filepaths, import_map, project=pm
        )
        return self._get_settings_from_version_map(files_by_solc_version, remappings, project=pm)

    def _get_settings_from_version_map(
        self,
        version_map: dict,
        import_remappings: dict[str, str],
        project: Optional[ProjectManager] = None,
        **kwargs,
    ) -> dict[Version, dict]:
        pm = project or self.local_project
        if not version_map:
            return {}

        config = self.get_config(project=pm)
        settings: dict = {}
        for solc_version, sources in version_map.items():
            version_settings: dict[str, Union[Any, list[Any]]] = {
                "optimizer": {"enabled": config.optimize, "runs": DEFAULT_OPTIMIZATION_RUNS},
                "outputSelection": {
                    str(get_relative_path(p, pm.path)): {"*": OUTPUT_SELECTION, "": ["ast"]}
                    for p in sorted(sources)
                },
                **kwargs,
            }
            if remappings_used := self._get_used_remappings(sources, import_remappings, project=pm):
                remappings_str = [f"{k}={v}" for k, v in remappings_used.items()]

                # Standard JSON input requires remappings to be sorted.
                version_settings["remappings"] = sorted(remappings_str)

            if evm_version := config.evm_version:
                version_settings["evmVersion"] = evm_version

            if solc_version >= Version("0.7.5") and config.via_ir is not None:
                version_settings["viaIR"] = config.via_ir

            settings[solc_version] = version_settings

            # TODO: Filter out libraries that are not used for this version.
            if libs := self.libraries:
                version_settings["libraries"] = libs

        return settings

    def _get_used_remappings(
        self,
        sources: Iterable[Path],
        remappings: dict[str, str],
        project: Optional[ProjectManager] = None,
    ) -> dict[str, str]:
        pm = project or self.local_project
        if not remappings:
            # No remappings used at all.
            return {}

        cache_path = (
            f"{get_relative_path(pm.contracts_folder.absolute(), pm.path)}{os.path.sep}.cache"
        )

        # Filter out unused import remapping.
        result = {}
        sources = list(sources)
        imports = self.get_imports(sources, project=pm).values()

        for source_list in imports:
            for src in source_list:
                if not src.startswith(cache_path):
                    continue

                parent_key = os.path.sep.join(src.split(os.path.sep)[:3])
                for k, v in remappings.items():
                    if parent_key in v:
                        result[k] = v

        return result

    def get_standard_input_json(
        self,
        contract_filepaths: Iterable[Path],
        project: Optional[ProjectManager] = None,
        **overrides,
    ) -> dict[Version, dict]:
        pm = project or self.local_project
        paths = list(contract_filepaths)  # Handle if given generator=
        remapping = self.get_import_remapping(project=pm)
        import_map = self.get_imports_from_remapping(paths, remapping, project=pm)
        version_map = self.get_version_map_from_imports(paths, import_map, project=pm)
        return self.get_standard_input_json_from_version_map(
            version_map, remapping, project=pm, **overrides
        )

    def get_standard_input_json_from(
        self,
        version_map: dict[Version, set[Path]],
        import_remappings: dict[str, str],
        project: Optional[ProjectManager] = None,
        **overrides,
    ):
        pm = project or self.local_project
        settings = self._get_settings_from_version_map(
            version_map, import_remappings, project=pm, **overrides
        )
        return self.get_standard_input_json_from_settings(settings, version_map, project=pm)

    def get_standard_input_json_from_version_map(
        self,
        version_map: dict[Version, set[Path]],
        import_remapping: dict[str, str],
        project: Optional[ProjectManager] = None,
        **overrides,
    ):
        pm = project or self.local_project
        settings = self._get_settings_from_version_map(
            version_map, import_remapping, project=pm, **overrides
        )
        return self.get_standard_input_json_from_settings(settings, version_map, project=pm)

    def get_standard_input_json_from_settings(
        self,
        settings: dict[Version, dict],
        version_map: dict[Version, set[Path]],
        project: Optional[ProjectManager] = None,
    ):
        pm = project or self.local_project
        input_jsons: dict[Version, dict] = {}

        for solc_version, vers_settings in settings.items():
            if not list(version_map[solc_version]):
                continue

            cleaned_version = Version(solc_version.base_version)
            solc_binary = get_executable(version=cleaned_version)
            arguments = {"solc_binary": solc_binary, "solc_version": cleaned_version}

            if solc_version >= Version("0.6.9"):
                arguments["base_path"] = pm.path

            if missing_sources := [
                x for x in vers_settings["outputSelection"] if not (pm.path / x).is_file()
            ]:
                # See if the missing sources are from dependencies (they likely are)
                # and cater the error message accordingly.
                if dependencies_needed := [x for x in missing_sources if str(x).startswith("@")]:
                    # Missing dependencies. Should only get here if dependencies are found
                    # in import-strs but are not installed anywhere (not in project or globally).
                    missing_str = ", ".join(dependencies_needed)
                    raise CompilerError(
                        f"Missing required dependencies '{missing_str}'. "
                        "Install them using `dependencies:` "
                        "in an ape-config.yaml or using the `ape pm install` command."
                    )

                # Otherwise, we are missing project-level source files for some reason.
                # This would only happen if the user passes in unexpected files outside
                # of core.
                missing_src_str = ", ".join(missing_sources)
                raise CompilerError(f"Sources '{missing_src_str}' not found in '{pm.name}'.")

            sources = {
                x: {"content": (pm.path / x).read_text()}
                for x in sorted(vers_settings["outputSelection"])
            }

            input_jsons[solc_version] = {
                "sources": sources,
                "settings": vers_settings,
                "language": "Solidity",
            }

        return {v: input_jsons[v] for v in sorted(input_jsons)}

    def compile(
        self,
        contract_filepaths: Iterable[Path],
        project: Optional[ProjectManager] = None,
        settings: Optional[dict] = None,
    ) -> Iterator[ContractType]:
        pm = project or self.local_project
        settings = settings or {}
        paths = [p for p in contract_filepaths]  # Handles generator.
        source_ids = [f"{get_relative_path(p.absolute(), pm.path)}" for p in paths]
        _validate_can_compile(paths)

        # Compile in an isolated env so the .cache folder does not interfere with anything.
        with pm.isolate_in_tempdir() as isolated_project:
            filepaths = [isolated_project.path / src_id for src_id in source_ids]
            yield from self._compile(filepaths, project=isolated_project, settings=settings)
            compilers = isolated_project.manifest.compilers

        pm.update_manifest(compilers=compilers)

    def _compile(
        self,
        contract_filepaths: Iterable[Path],
        project: Optional[ProjectManager] = None,
        settings: Optional[dict] = None,
    ):
        pm = project or self.local_project
        input_jsons = self.get_standard_input_json(
            contract_filepaths, project=pm, **(settings or {})
        )
        contract_versions: dict[str, Version] = {}
        contract_types: list[ContractType] = []
        for solc_version, input_json in input_jsons.items():
            keys = (
                "\n\t".join(sorted([x for x in input_json.get("sources", {}).keys()]))
                or "No input."
            )
            log_str = f"Compiling using Solidity compiler '{solc_version}'.\nInput:\n\t{keys}"
            logger.info(log_str)
            cleaned_version = Version(solc_version.base_version)
            solc_binary = get_executable(version=cleaned_version)
            arguments: dict = {"solc_binary": solc_binary, "solc_version": cleaned_version}

            if solc_version >= Version("0.6.9"):
                arguments["base_path"] = pm.path

            # Allow empty contracts, like Vyper does.
            arguments["allow_empty"] = True

            try:
                output = compile_standard(input_json, **arguments)
            except SolcError as err:
                raise SolcCompileError(err) from err

            contracts = output.get("contracts", {})
            # Perf back-out.
            if not contracts:
                continue

            input_contract_names: list[str] = []
            for source_id, contracts_out in contracts.items():
                for name, _ in contracts_out.items():
                    # Filter source files that the user did not ask for, such as
                    # imported relative files that are not part of the input.
                    for input_file_path in contract_filepaths:
                        if source_id in str(input_file_path):
                            input_contract_names.append(name)

            for source_id, contracts_out in contracts.items():
                # ast_data = output["sources"][source_id]["ast"]

                for contract_name, ct_data in contracts_out.items():
                    if contract_name not in input_contract_names:
                        # Only return ContractTypes explicitly asked for.
                        continue

                    evm_data = ct_data["evm"]

                    # NOTE: This sounds backwards, but it isn't...
                    #  The "deployment_bytecode" is the same as the "bytecode",
                    #  and the "deployedBytecode" is the same as the "runtimeBytecode".
                    deployment_bytecode = add_0x_prefix(evm_data["bytecode"]["object"])
                    runtime_bytecode = add_0x_prefix(evm_data["deployedBytecode"]["object"])

                    # Skip library linking.
                    if "__$" in deployment_bytecode or "__$" in runtime_bytecode:
                        logger.warning(
                            f"Unable to compile {contract_name} - missing libraries. "
                            f"Call `{self.add_library.__name__}` with the necessary libraries"
                        )
                        continue

                    if contract_name in contract_versions:
                        # Already yield in smaller version. Must not yield again
                        # or else we will have a contract-type collision.
                        # (Sources that are required in multiple version-sets will
                        # hit this).
                        continue

                    ct_data["contractName"] = contract_name
                    ct_data["sourceId"] = source_id
                    ct_data["deploymentBytecode"] = {"bytecode": deployment_bytecode}
                    ct_data["runtimeBytecode"] = {"bytecode": runtime_bytecode}
                    ct_data["userdoc"] = load_dict(ct_data["userdoc"])
                    ct_data["devdoc"] = load_dict(ct_data["devdoc"])
                    ct_data["sourcemap"] = evm_data["bytecode"]["sourceMap"]
                    contract_type = ContractType.model_validate(ct_data)
                    yield contract_type
                    contract_types.append(contract_type)
                    contract_versions[contract_name] = solc_version

        # Output compiler data used.
        compilers_used: dict[Version, Compiler] = {}
        for ct in contract_types:
            if not ct.name:
                # Won't happen, but just for mypy.
                continue

            vers = contract_versions[ct.name]
            settings = input_jsons[vers]["settings"]
            if vers in compilers_used and ct.name not in (compilers_used[vers].contractTypes or []):
                compilers_used[vers].contractTypes = [
                    *(compilers_used[vers].contractTypes or []),
                    ct.name,
                ]

            elif vers not in compilers_used:
                compilers_used[vers] = Compiler(
                    name=self.name.lower(),
                    version=f"{vers}",
                    contractTypes=[ct.name],
                    settings=settings,
                )

        # Update compilers used in project manifest.
        # First, output compiler information to manifest.
        compilers_ls = list(compilers_used.values())
        pm.add_compiler_data(compilers_ls)

    def compile_code(
        self,
        code: str,
        project: Optional[ProjectManager] = None,
        **kwargs,
    ) -> ContractType:
        pm = project or self.local_project
        if settings_version := self._get_configured_version(project=pm):
            version = settings_version

        elif pragma := self._get_pramga_spec_from_str(code):
            if selected_version := select_version(pragma, self.installed_versions):
                version = selected_version
            else:
                if selected_version := select_version(pragma, self.available_versions):
                    version = selected_version
                    install_solc(version, show_progress=True)
                else:
                    raise SolcInstallError()

        elif latest_installed := self.latest_installed_version:
            version = latest_installed

        elif latest := self.latest_version:
            install_solc(latest, show_progress=True)
            version = latest

        else:
            raise SolcInstallError()

        version = add_commit_hash(version)
        cleaned_version = Version(version.base_version)
        executable = get_executable(cleaned_version)
        try:
            result = compile_source(
                code,
                import_remappings=self.get_import_remapping(project=pm),
                base_path=pm.path,
                solc_binary=executable,
                solc_version=cleaned_version,
                allow_empty=True,
            )
        except SolcError as err:
            raise SolcCompileError(err) from err

        output = result[next(iter(result.keys()))]
        return ContractType.model_validate(
            {
                "abi": output["abi"],
                "ast": output["ast"],
                "deploymentBytecode": {"bytecode": HexBytes(output["bin"])},
                "devdoc": load_dict(output["devdoc"]),
                "runtimeBytecode": {"bytecode": HexBytes(output["bin-runtime"])},
                "sourcemap": output["srcmap"],
                "userdoc": load_dict(output["userdoc"]),
                **kwargs,
            }
        )

    def get_imports(
        self,
        contract_filepaths: Iterable[Path],
        project: Optional[ProjectManager] = None,
    ) -> dict[str, list[str]]:
        pm = project or self.local_project
        remapping = self.get_import_remapping(project=pm)
        _validate_can_compile(contract_filepaths)
        paths = [x for x in contract_filepaths]  # Handle if given generator.
        return self.get_imports_from_remapping(paths, remapping, project=pm)

    def get_imports_from_remapping(
        self,
        paths: Iterable[Path],
        remapping: dict[str, str],
        project: Optional[ProjectManager] = None,
    ) -> dict[str, list[str]]:
        pm = project or self.local_project
        return self._get_imports(paths, remapping, pm, tracked=set())  # type: ignore

    def _get_imports(
        self,
        paths: Iterable[Path],
        remapping: dict[str, str],
        pm: "ProjectManager",
        tracked: set[str],
        include_raw: bool = False,
    ) -> dict[str, Union[dict[str, str], list[str]]]:
        result: dict = {}

        for src_path, import_strs in get_import_lines(paths).items():
            source_id = str(get_relative_path(src_path, pm.path))
            if source_id in tracked:
                # We have already accumulated imports from this source.
                continue

            tracked.add(source_id)

            # Init with all top-level imports.
            import_map = {
                x: self._import_str_to_source_id(x, src_path, remapping, project=pm)
                for x in import_strs
            }
            import_source_ids = list(set(list(import_map.values())))

            # NOTE: Add entry even if empty here.
            result[source_id] = import_map if include_raw else import_source_ids

            # Add imports of imports.
            if not result[source_id]:
                # Nothing else imported.
                continue

            # Add known imports.
            known_imports = {p: result[p] for p in import_source_ids if p in result}
            imp_paths = [pm.path / p for p in import_source_ids if p not in result]
            unknown_imports = self._get_imports(
                imp_paths,
                remapping,
                pm,
                tracked=tracked,
                include_raw=include_raw,
            )
            sub_imports = {**known_imports, **unknown_imports}

            # All imported sources from imported sources are imported sources.
            for sub_set in sub_imports.values():
                if isinstance(sub_set, dict):
                    for import_str, sub_import in sub_set.items():
                        result[source_id][import_str] = sub_import

                else:
                    for sub_import in sub_set:
                        if sub_import not in result[source_id]:
                            result[source_id].append(sub_import)

                    # Keep sorted.
                    result[source_id] = sorted((result[source_id]))

            # Combine results. This ends up like a tree-structure.
            result = {**result, **sub_imports}

        # Sort final keys and import lists for more predictable compiler behavior.
        return {k: result[k] for k in sorted(result.keys())}

    def get_version_map(
        self,
        contract_filepaths: Union[Path, Iterable[Path]],
        project: Optional[ProjectManager] = None,
    ) -> dict[Version, set[Path]]:
        pm = project or self.local_project
        paths = (
            [contract_filepaths]
            if isinstance(contract_filepaths, Path)
            else [p for p in contract_filepaths]
        )
        _validate_can_compile(paths)
        imports = self.get_imports(paths, project=pm)
        return self.get_version_map_from_imports(paths, imports, project=pm)

    def get_version_map_from_imports(
        self,
        contract_filepaths: Union[Path, Iterable[Path]],
        import_map: dict[str, list[str]],
        project: Optional[ProjectManager] = None,
    ):
        pm = project or self.local_project
        paths = (
            [contract_filepaths]
            if isinstance(contract_filepaths, Path)
            else [p for p in contract_filepaths]
        )
        path_set: set[Path] = {p for p in paths}

        # Add imported source files to list of contracts to compile.
        for source_path in paths:
            source_id = f"{get_relative_path(source_path, pm.path)}"
            if source_id not in import_map or len(import_map[source_id]) == 0:
                continue

            import_set = {pm.path / src_id for src_id in import_map[source_id]}
            path_set = path_set.union(import_set)

        # Use specified version if given one
        if _version := self._get_configured_version(project=pm):
            return {_version: path_set}
        # else: find best version per source file

        # Build map of pragma-specs.
        pragma_map = {p: get_pragma_spec_from_path(p) for p in path_set}

        # If no Solidity version has been installed previously while fetching the
        # contract version pragma, we must install a compiler, so choose the latest
        if (
            not self.installed_versions
            and not any(pragma_map.values())
            and (latest := self.latest_version)
        ):
            install_solc(latest, show_progress=True)

        # Adjust best-versions based on imports.
        files_by_solc_version: dict[Version, set[Path]] = {}
        for source_file_path in path_set:
            solc_version = self._get_best_version(source_file_path, pragma_map)
            imported_source_paths = self._get_imported_source_paths(
                source_file_path, pm.path, import_map
            )

            for imported_source_path in imported_source_paths:
                if imported_source_path not in pragma_map:
                    continue

                imported_pragma_spec = pragma_map[imported_source_path]
                imported_version = self._get_best_version(imported_source_path, pragma_map)

                if imported_pragma_spec is not None and (
                    str(imported_pragma_spec)[0].startswith("=")
                    or str(imported_pragma_spec)[0].isdigit()
                ):
                    # Have to use this version.
                    solc_version = imported_version
                    break

                elif imported_version < solc_version:
                    # If we get here, the highest version of an import is lower than the reference.
                    solc_version = imported_version

            if solc_version not in files_by_solc_version:
                files_by_solc_version[solc_version] = set()

            for path in (source_file_path, *imported_source_paths):
                files_by_solc_version[solc_version].add(path)

        # If being used in another version AND no imports in this version require it,
        # remove it from this version.
        for solc_version, files in files_by_solc_version.copy().items():
            for file in files.copy():
                used_in_other_version = any(
                    [file in ls for v, ls in files_by_solc_version.items() if v != solc_version]
                )
                if not used_in_other_version:
                    continue

                other_files = [f for f in files_by_solc_version[solc_version] if f != file]
                used_in_imports = False
                for other_file in other_files:
                    source_id = str(get_relative_path(other_file, pm.path))
                    import_paths = [pm.path / i for i in import_map.get(source_id, []) if i]
                    if file in import_paths:
                        used_in_imports = True
                        break

                if not used_in_imports:
                    files_by_solc_version[solc_version].remove(file)
                    if not files_by_solc_version[solc_version]:
                        del files_by_solc_version[solc_version]

        result = {add_commit_hash(v): ls for v, ls in files_by_solc_version.items()}

        # Sort, so it is a nicer version map and the rest of the compilation flow
        # is more predictable.
        return {k: result[k] for k in sorted(result)}

    def _get_imported_source_paths(
        self,
        path: Path,
        base_path: Path,
        imports: dict,
        source_ids_checked: Optional[list[str]] = None,
    ) -> set[Path]:
        source_ids_checked = source_ids_checked or []
        source_identifier = str(get_relative_path(path, base_path))
        if source_identifier in source_ids_checked:
            # Already got this source's imports
            return set()

        source_ids_checked.append(source_identifier)
        import_file_paths = [base_path / i for i in imports.get(source_identifier, []) if i]
        return_set = {i for i in import_file_paths}
        for import_path in import_file_paths:
            indirect_imports = self._get_imported_source_paths(
                import_path, base_path, imports, source_ids_checked=source_ids_checked
            )
            for indirect_import in indirect_imports:
                return_set.add(indirect_import)

        return return_set

    def _get_pramga_spec_from_str(self, source_str: str) -> Optional[SpecifierSet]:
        if not (pragma_spec := get_pragma_spec_from_str(source_str)):
            return None

        # Check if we need to install specified compiler version
        if select_version(pragma_spec, self.installed_versions):
            return pragma_spec

        elif compiler_version := select_version(pragma_spec, self.available_versions):
            install_solc(compiler_version, show_progress=True)

        else:
            # Attempt to use the best-installed version.
            for _version in self.installed_versions:
                if _version not in pragma_spec:
                    continue

                logger.warning(
                    "Ape is unable to determine if additional versions are needed "
                    f"in order to meet spec {pragma_spec}. Resorting to the best matching "
                    "already-installed version. Alternatively, specify a Solidity "
                    "compiler version in your ape-config.yaml."
                )
                return pragma_spec

            # None of the installed versions match, and we are unable to install.
            raise CompilerError(f"Solidity version specification '{pragma_spec}' could not be met.")

        return pragma_spec

    def _get_best_versions(self, path: Path, options, source_by_pragma_spec: dict) -> list[Version]:
        # NOTE: Doesn't install.
        if pragma_spec := source_by_pragma_spec.get(path):
            res = get_versions_can_use(pragma_spec, list(options))
        elif latest_installed := self.latest_installed_version:
            res = [latest_installed]
        elif latest := self.latest_version:
            res = [latest]
        else:
            raise SolcInstallError()

        return [add_commit_hash(v) for v in res]

    def _get_best_version(self, path: Path, source_by_pragma_spec: dict) -> Version:
        compiler_version: Optional[Version] = None
        if pragma_spec := source_by_pragma_spec.get(path):
            if selected := select_version(pragma_spec, self.installed_versions):
                compiler_version = selected

            elif selected := select_version(pragma_spec, self.available_versions):
                # Install missing version.
                # NOTE: Must be installed before adding commit hash.
                install_solc(selected, show_progress=True)
                compiler_version = add_commit_hash(selected)

        elif latest_installed := self.latest_installed_version:
            compiler_version = latest_installed

        elif latest := self.latest_version:
            # Download latest version.
            install_solc(latest, show_progress=True)
            compiler_version = latest

        else:
            raise SolcInstallError()

        assert compiler_version  # For mypy
        return add_commit_hash(compiler_version)

    def enrich_error(self, err: ContractLogicError) -> ContractLogicError:
        if not is_0x_prefixed(err.revert_message):
            # Nothing to do.
            return err

        elif panic_cls := _get_sol_panic(err.revert_message):
            return panic_cls(
                base_err=err.base_err,
                contract_address=err.contract_address,
                source_traceback=err.source_traceback,
                trace=err.trace,
                txn=err.txn,
            )

        # Check for ErrorABI.
        bytes_message = HexBytes(err.revert_message)
        selector = bytes_message[:4]
        input_data = bytes_message[4:]

        if not err.address:
            return err

        if not self.network_manager.active_provider:
            # Connection required.
            return err

        if (
            not (contract := self.chain_manager.contracts.instance_at(err.address))
            or selector not in contract.contract_type.errors
        ):
            return err

        ecosystem = self.provider.network.ecosystem
        abi = contract.contract_type.errors[selector]
        inputs = ecosystem.decode_calldata(abi, input_data)
        error_class = contract.get_error_by_signature(abi.signature)
        return error_class(
            abi,
            inputs,
            base_err=err.base_err,
            contract_address=err.contract_address,
            source_traceback=err.source_traceback,
            trace=err.trace,
            txn=err.txn,
        )

    def _flatten_source(
        self,
        path: Path,
        project: Optional[ProjectManager] = None,
        raw_import_name: Optional[str] = None,
        handled: Optional[set[str]] = None,
    ) -> str:
        pm = project or self.local_project
        handled = handled or set()
        source_id = f"{get_relative_path(path, pm.path)}"
        handled.add(source_id)
        remapping = self.get_import_remapping(project=project)
        imports = self._get_imports((path,), remapping, pm, tracked=set(), include_raw=True)
        relevant_imports = imports.get(source_id, {})

        final_source = ""

        for import_str, source_id in relevant_imports.items():  # type: ignore
            if source_id in handled:
                continue

            sub_import_name = import_str.replace("import ", "").strip(" \n\t;\"'")
            final_source += self._flatten_source(
                pm.path / source_id,
                project=pm,
                raw_import_name=sub_import_name,
                handled=handled,
            )

        final_source += _get_flattened_source(path, name=raw_import_name)
        return final_source

    def flatten_contract(
        self, path: Path, project: Optional[ProjectManager] = None, **kwargs
    ) -> Content:
        # try compiling in order to validate it works
        res = self._flatten_source(path, project=project)
        res = remove_imports(res)
        res = process_licenses(res)
        res = remove_version_pragmas(res)
        pragma = get_first_version_pragma(path.read_text())
        res = "\n".join([pragma, res])
        lines = res.splitlines()
        line_dict = {i + 1: line for i, line in enumerate(lines)}
        return Content(root=line_dict)

    def _import_str_to_source_id(
        self,
        _import_str: str,
        source_path: Path,
        import_remapping: dict[str, str],
        project: Optional[ProjectManager] = None,
    ) -> str:
        pm = project or self.local_project
        quote = '"' if '"' in _import_str else "'"

        try:
            end_index = _import_str.index(quote) + 1
        except ValueError as err:
            raise CompilerError(
                f"Error parsing import statement '{_import_str}' in '{source_path.name}'."
            ) from err

        import_str_prefix = _import_str[end_index:]
        import_str_value = import_str_prefix[: import_str_prefix.index(quote)]

        # Get all matches.
        valid_matches: list[tuple[str, str]] = []
        key = None
        base_path = None
        for key, value in import_remapping.items():
            if key not in import_str_value:
                continue

            valid_matches.append((key, value))

        if valid_matches:
            key, value = max(valid_matches, key=lambda x: len(x[0]))
            import_str_value = import_str_value.replace(key, value)

        if import_str_value.startswith("."):
            base_path = source_path.parent
        elif (pm.path / import_str_value).is_file():
            base_path = pm.path
        elif (pm.contracts_folder / import_str_value).is_file():
            base_path = pm.contracts_folder
        elif key is not None and key.startswith("@"):
            nm = key[1:]
            for cfg_dep in pm.config.dependencies:
                if (
                    cfg_dep.get("name") == nm
                    and "project" in cfg_dep
                    and (Path(cfg_dep["project"]) / import_str_value).is_file()
                ):
                    base_path = Path(cfg_dep["project"])

        if base_path is None:
            # No base_path, do as-is.
            return import_str_value

        path = (base_path / import_str_value).resolve()
        return f"{get_relative_path(path.absolute(), pm.path)}"


def remove_imports(flattened_contract: str) -> str:
    # Use regex.sub() to remove matched import statements
    no_imports_contract = IMPORTS_PATTERN.sub("", flattened_contract)

    return no_imports_contract


def remove_version_pragmas(flattened_contract: str) -> str:
    return VERSION_PRAGMA_PATTERN.sub("", flattened_contract)


def get_first_version_pragma(source: str) -> str:
    match = VERSION_PRAGMA_PATTERN.search(source)
    if match:
        return match.group(0)
    return ""


def get_licenses(src: str) -> list[tuple[str, str]]:
    return LICENSES_PATTERN.findall(src)


def process_licenses(contract: str) -> str:
    """
    Process the licenses in a contract.
    Ensure that all licenses are identical, and if not, raise an error.
    The contract is returned with a single license identifier line at its top.
    """

    # Extract SPDX license identifiers from the contract.
    extracted_licenses = get_licenses(contract)

    # Return early if no licenses are present in the contract.
    if not extracted_licenses:
        return contract

    # The "root" license is most-likely last because import statements are
    # replaced with their sources typically at the top of the file
    license_line, root_license = extracted_licenses[-1]

    # Get the unique license identifiers. All licenses in a contract _should_ be the same.
    unique_license_identifiers = {
        license_identifier for _, license_identifier in extracted_licenses
    }

    # If we have more than one unique license identifier, warn the user and use the root.
    if len(unique_license_identifiers) > 1:
        licenses_str = ", ".join(sorted(unique_license_identifiers))
        logger.warning(
            f"Conflicting licenses found: '{licenses_str}'. "
            f"Using the root file's license '{root_license}'."
        )

    # Remove all of the license lines from the contract.
    contract_without_licenses = contract
    for license_tuple in extracted_licenses:
        line = license_tuple[0]
        contract_without_licenses = contract_without_licenses.replace(line, "")

    # Prepend the contract with only the root license line.
    contract_with_single_license = f"{license_line}\n{contract_without_licenses}"

    return contract_with_single_license


def _get_sol_panic(revert_message: str) -> Optional[type[RuntimeErrorUnion]]:
    if revert_message.startswith(RUNTIME_ERROR_CODE_PREFIX):
        # ape-geth (style) plugins show the hex with the Panic ABI prefix.
        error_type_val = int(
            f"0x{revert_message.replace(RUNTIME_ERROR_CODE_PREFIX, '').lstrip('0')}", 16
        )
    else:
        # Some plugins, like ape-hardhat, will deliver panic codes directly (no Panic ABI prefix)
        error_type_val = int(revert_message, 16)

    if error_type_val in [x.value for x in RuntimeErrorType]:
        return RUNTIME_ERROR_MAP[RuntimeErrorType(error_type_val)]

    return None


def _try_max(ls: list[Any]):
    return max(ls) if ls else None


def _validate_can_compile(paths: Iterable[Path]):
    extensions = {get_full_extension(p): p for p in paths if p}

    for ext, file in extensions.items():
        if ext not in [e.value for e in Extension]:
            raise CompilerError(f"Unable to compile '{file.name}' using Solidity compiler.")
